%{
  /* basico02.l
   * Tiny BASIC no GOTO lexer
   *
   * by Andre Adrian
   *
   * version 04jun2006
   */ 
  #include "y.tab.h"
  #include <stdlib.h>
  void yyerror(char *);
  
  extern int namespace; /* import from symtab.c */
  extern char ident[];  /* import from codegen.c */
%}

%%

"array"     { return ARRAY; }
"begin"     { return BEGINN; }
"break"     { return BREAK; }
"char"      { return CHAR; }
"do"	    { return DO; }
"else"	    { return ELSE; }
"end"       { return END; }
"endif"	    { return ENDIF; }
"forward"   { return FORWARD; }
"func"      { return FUNC; }
"if"  	    { return IF; }
"int"       { return INT; }
"return"    { return RETURN; }
"then"	    { return THEN; }
"var"       { return VAR; }
"void"      { return VOID; }
"wend"	    { return WEND; }
"while"	    { return WHILE; }
"<="	    { return LE; }
">="	    { return GE; }


'.'         { yylval = yytext[1];
              return CHRCONST;
            }
            
\".+\"      { yylval = yytext[1];
              strncpy(ident, yytext, 127);
              return STRCONST;
            }
            
[A-Za-z][A-Za-z0-9]*	{ yylval = yytext[0];
      	      strncpy(ident, yytext, 31);
              return IDENT;
            }

[0-9]+      { yylval = atoi(yytext);
              return CONST;
            }

[-+*/,:;()=#<>&|!%]  { return *yytext; }

"["         { return *yytext; }
"]"         { return *yytext; }

^[ \t]*\n   ; 	    /* skip empty line */
\\[ \t]*\n  ; 	    /* skip line continuation */

\n          { return ';'; }

"//".*\n    ; 	    /* skip one-line comment */

[ \t]       ;       /* skip whitespace */

.           yyerror("Unknown character");

%%

int yywrap(void) {
    return 1;
}
