%{
  /* basico02.y
   * Tiny BASIC no GOTO parser
   *
   * by Andre Adrian
   *
   * version 04jun2006
   */ 
  #include <stdio.h>
  #include "codegen.h"
  int yylex(void);
  void yyerror(char *);

%}

%token  CONST CHRCONST STRCONST IDENT
%token	VAR INT CHAR ARRAY VOID
%token	FUNC FORWARD BEGINN END RETURN
%token	IF THEN ELSE ENDIF
%token	WHILE DO BREAK WEND
%token  LE GE

%%

/* based on Pascal BNF */

program:  
   block4  
   | variable_declaration ';' block4
   ;
   
block4:  
   func_declaration ';'
   ;
   
variable_declaration:  
   VAR global_decl  
   | variable_declaration ';' global_decl
   ;
   
global_decl:
   declid ':' INT                 { declint(1); }                      
   | declid ':' ARRAY CONST INT   { declint($4);  }         
   | declid ':' CHAR              { declchar(1); }                     
   | declid ':' ARRAY CONST CHAR  { declchar($4);  }
   ;

declid:
   IDENT                            { declid(); }

func_declaration:
   func
   | func_declaration ';' func
   ;

func:
   FUNC funcid '(' param_list ')' ':' return_type ';' block_or_forward
   ;

funcid:
   IDENT                    { funcid(); }
   ;

param_list:
   IDENT ':' param_type
   | param_list ',' IDENT ':' param_type
   |  	 /* empty set */
   ;

param_type:
   INT                           { printf("#param %c int\n", $1); }        
   | CHAR                        { printf("#param %c char\n", $1); }       
   | ARRAY INT                   { printf("#param %c int_array\n", $1); }  
   | ARRAY CHAR                  { printf("#param %c char_array\n", $1); } 
   ;
      
return_type:
   INT                           { printf("#func returns int\n"); }        
   | CHAR                        { printf("#func returns char\n"); }       
   | VOID                        { printf("#func returns void\n"); }       
   ;
   
block_or_forward:  
   pblock3                       
   | FORWARD                     { printf("#forward decl\n");  addglobals(); }            
   ;
   
pblock3:  
   pblock4  
   | pvariable_declaration ';' pblock4
   ;
   
pblock4:  
   beginl stmt_list END          { funcepilogue(); }
   ;
   
beginl:
   BEGINN                        { funcprologue(0); }
   ;
   
pvariable_declaration:  
   VAR local_decl  
   | pvariable_declaration ';' local_decl
   ;
   
local_decl:
   IDENT ':' INT                 { printf("#local %c int\n", $1); }                       
   | IDENT ':' CHAR              { printf("#local %c char\n", $1); }                      
   | IDENT ':' ARRAY CONST INT   { printf("#local %c int_array %d\n", $1, $4); }          
   | IDENT ':' ARRAY CONST CHAR  { printf("#local %c char_array %d\n", $1, $4); }         
   ;

stmt_list:  
   statement ';'
   | stmt_list statement ';'
   ;  

statement:
   variable '=' expr             { /* use type */ assign(); }
   | IF expr thenl stmt_list else_or_endif
   | whilel expr dol stmt_list WEND   { wendjump(); }
   | callid '(' expr_list ')'    { callproc(); }
   | RETURN                      { retrn(); }
   | RETURN expr                 { retrnvalue(); }
   | BREAK                       { breakjump(); }
   |   	 /* empty set */
   ;
   
variable:
   varid                     
   | varid '[' expr ']'           { /* save type */ assignarray(); }   
   ;  

varid:
   IDENT                          { varid(); }
    
whilel:
   WHILE             	      	  { whilelabel(); }
   ;

dol:
   DO             	      	  { dojump(); }
   ;

thenl:
   THEN                           { thenjump(); }

else_or_endif:
   ENDIF                          { endiflabel(); }
   | elsel stmt_list ENDIF        { endiflabel(); }
   ;
   
elsel:
   ELSE                      { elsejump(); }
   ;

callid:
   IDENT                     { callid(); }
   ;
   
expr_list:
   expr                      { /* C pushes from right to left */ callarg(); }                    
   | expr_list ',' expr      { callarg(); }                    
   |  	 /* empty set */
   ;
  
expr:
   addexpr
   | expr '=' addexpr       { eq(); }                                      
   | expr '#' addexpr       { ne(); }                                      
   | expr '<' addexpr       { lt(); }                                      
   | expr '>' addexpr       { gt(); }                                      
   | expr LE  addexpr       { le(); }    
   | expr GE  addexpr       { ge(); }    
   ;

addexpr:
   term
   | addexpr '+' term       { add(); }     
   | addexpr '-' term       { sub(); }    
   | addexpr '|' term       { or(); }     
   ;

term:
   unary_fact
   | term '*' unary_fact     { mul(); }                      
   | term '/' unary_fact     { div(); }                      
   | term '%' unary_fact     { mod(); }                      
   | term '&' unary_fact     { and(); }                      
   ;

unary_fact: 
   '-' unary_fact            { neg(); }                          
   | '!' unary_fact          { not(); }                          
   | fact
   ;

fact:
   CONST      	      	      { pushconst($1); }                     
   | CHRCONST      	      { pushchrconst($1); }                     
   | STRCONST      	      { pushstrconst(); }                     
   | IDENT                    { pushvar(); }                       
   | arrayid '[' expr ']'     { pusharrayvar(); }            
   | '(' expr ')'
   | callid '(' expr_list ')' { callfunc(); }             
   ;
   
arrayid:
   IDENT                      { arrayid(); }
   ;   

%%

void yyerror(char *s) {
    fprintf(stderr, "%s\n", s);
}

int main(void) {
    addglobals();
    fileprologue();
#if YYDEBUG
    yydebug = 1;
#endif    
    yyparse();
    fileepilogue();
    return 0;
}
