/* codegen.c
 * GNU as / Intel 386 32bit mode Code Generator 
 * Input: Stack-machine Opcodes
 * Output: Opcodes for Registers as Stack cache with Register renaming
 *
 * by Andre Adrian
 *
 * version 04jun2006
 *
 * compiler bug: srand(time(0)) not working
 * nested funcs need stacked oldse?
 */

#include <stdio.h>

// uncomment to test stand-alone
// #define MODULTEST

#define REGS 3          // bx, ax, cx
#define LEVELS 10       // if nesting level

int se;                 // Stack elements
int oldse;
char reg[REGS];         // Register usage

int label;    	      	// Label-id generator
int L1[LEVELS], L2[LEVELS]; // if jump label stack
int L3[LEVELS], L4[LEVELS]; // while jump label stack
int level;              // for if
int whilelevel;         // for while 
int lco;                // String const id
char ident[128];
char iddecl[32];        // variable definition under construction
char idfunc[32];        // function definition under construction
char idcall[32];        // function call under construction
char idvar[32];         // assignment under construction
char idarray[32];       // variable access under construction
int args;      
int isarray;

void cacheinit() {
  se = 0;
  reg[0] = 'c';
  reg[1] = 'b';
  reg[2] = 'a';
}

void fileprologue() {
  cacheinit();
  ident[127] = '\0';
  label = 0;
  level = -1;
  whilelevel = -1;
  lco = 0;
  printf("\t.file  \"a.bas\"\n");
  printf("\t.text\n");
}

void fileepilogue() {
  printf("\t.section .note.GNU-stack,\"\",@progbits\n");
  printf("\t.ident \"Basico 0.2\"\n");
}

void declid() {
  strcpy(iddecl, ident);
}  

void declchar(int elem) {
  printf("\t.comm  %s,%d,1\n", iddecl, elem);
}

void declint(int elem) {
  printf("\t.comm  %s,%d,4\n", iddecl, elem*4);
}


void funcid() {
  printf("\t\t#func decl\n");
  strcpy(idfunc, ident);
}

void funcprologue(int stackspace) {
  printf(".globl\t%s\n", idfunc);
  printf("\t.type\t%s, @function\n", idfunc);
  printf("%s:\n", idfunc);
  printf("\tpushl\t%%ebp\n");
  printf("\tmovl\t%%esp, %%ebp\n");
  if (stackspace > 0) {
    printf("\tsubl\t$%d, %%esp\n", stackspace);
  }
}

void funcepilogue() {
  printf("\tleave\n");
  printf("\tret\n");
  printf("\t.size %s, .-%s\n", idfunc, idfunc);
}

void retrn() {
  printf("\tleave\n");
  printf("\tret\n");
}

void retrnvalue() {
  // return value in reg eax
  printf("\t\t#pop to return-reg, return\n");
  if (reg[0] != 'a') {
    printf("\tmovl   %%e%cx, %%eax\n", reg[0]);
  }
  retrn();
}

pushcache() {
  if (se >= 3) {
    printf("\tpushl\t%%e%cx\n", reg[2]);
  }
  if (se >= 2) {
    printf("\tpushl\t%%e%cx\n", reg[1]);
  }
  if (se >= 1) {
    printf("\tpushl\t%%e%cx\n", reg[0]);
  }
}

popcache() {
  if (se >= 2) {
    printf("\tpopl\t%%e%cx\n", reg[1]);
  }
  if (se >= 3) {
    printf("\tpopl\t%%e%cx\n", reg[2]);
  }
}

pushargs() {
  /* C pushes from right to left */
  if (se == 3) {
    printf("\tpushl\t%%e%cx\n", reg[0]);
    printf("\tpushl\t%%e%cx\n", reg[1]);
    printf("\tpushl\t%%e%cx\n", reg[2]);
  }
  if (se == 2) {
    printf("\tpushl\t%%e%cx\n", reg[0]);
    printf("\tpushl\t%%e%cx\n", reg[1]);
  }
  if (se == 1) {
    printf("\tpushl\t%%e%cx\n", reg[0]);
  }
}

void callid() {
  pushcache();
  oldse = se;
  se = 0;
  printf("\tsubl\t$12, %%esp");
  printf("\t#callid\n");
  strcpy(idcall, ident);
  args = 0;
}
  
void callarg() {
  printf("\t\t#callarg\n");
  ++args;
  if (args > 3) {
    fprintf(stderr, "only 3 args\n");
    exit(1);
  }
}

void call() {
  printf("\t\t\t\t#call\n");
  pushargs();
  printf("\tcall\t%s\n", idcall);
  printf("\taddl\t$%d, %%esp", 12+4*args);
  printf("\t#drop args\n");
}

void callproc() {
  call();
  cacheinit();
}

void callfunc() {
  call();
  printf("\t\t#push return value\n");
  se = oldse+1;
  reg[0] = 'a';   // return value in eax
  reg[1] = 'b';
  reg[2] = 'c';
  popcache();
}

void thenjump() {
  // L1, L2 are local variables in a recursive decent parser
  ++level;
  L1[level] = label++;
  L2[level] = L1[level];
  printf("\tandl\t%%e%cx, %%e%cx", reg[0], reg[0]);
  printf("\t#then_jump %d\n", level);
  printf("\tjz\t.L%d\n", L1[level]);
  cacheinit();
}

void elsejump() {
  L2[level] = label++;
  printf("\tjmp\t.L%d", L2[level]);
  printf("\t\t#else_jump %d\n", level); 
  printf(".L%d:\n", L1[level]);
}

void endiflabel() {
  printf(".L%d:", L2[level]);
  printf("\t\t#endif_label %d\n", level);
  --level;
}

void whilelabel() {
  // L3, L4 are local variables in a recursive decent parser
  ++whilelevel;
  L3[whilelevel] = label++;
  L4[whilelevel] = label++;
  printf(".L%d:", L3[whilelevel]);
  printf("\t#while_label %d\n", whilelevel);
}

void dojump() {
  // L3, L4 are local variables in a recursive decent parser
  printf("\tandl\t%%e%cx, %%e%cx", reg[0], reg[0]);
  printf("\t#do_jump %d\n", whilelevel);
  printf("\tjz\t.L%d\n", L4[whilelevel]);
  cacheinit();
}

void breakjump() {
  if (whilelevel < 0) {
    fprintf(stderr, "break without while\n");
    exit(1);
  }
  printf("\tjmp\t.L%d", L4[whilelevel]);
  printf("\t#break_jump %d\n", whilelevel);
}
  
void wendjump() {
  printf("\tjmp\t.L%d", L3[whilelevel]);
  printf("\t\t#wend_jump %d\n", whilelevel); 
  printf(".L%d:\n", L4[whilelevel]);
  --whilelevel;
}

void push() {
  char tmp;
  if (se >= REGS) {
    printf("\tpushl\t%%e%cx\n", reg[2]);
  }
  tmp = reg[2];
  reg[2] = reg[1];
  reg[1] = reg[0];
  reg[0] = tmp;
  ++se;
}

void pop() {
  char tmp;
  --se;
  if (se == 0) {
    cacheinit();
  } else {
    tmp = reg[0]; 
    reg[0] = reg[1];
    reg[1] = reg[2];
    reg[2] = tmp;
  }
  if (se >= REGS) {
    printf("\tpopl\t%%e%cx\n", reg[2]);
  }
}

void arrayid() {
  strcpy(idarray, ident);
}  

varid() {
  isarray = 0;
  printf("\t\t\t\t#varid %s\n", ident);
  strcpy(idvar, ident);
}

assignarray() {
  isarray = 1;
}

assign() {
  // type and array dependent
  if (1 == isarray) {
    printf("\tmovl\t%%e%cx, %s(,%%e%cx,4)", reg[0], idvar, reg[1]);
    pop();    
  } else {
    printf("\tmovl\t%%e%cx, %s", reg[0], idvar);
  }
  printf("\t#assign\n");
  pop();
}

pushconst(int value) {
  push();
  printf("\tmovl\t$%d, %%e%cx", value, reg[0]);
  printf("\t#pushconst %d\n", value);
}

pushchrconst(int value) {
  push();
  printf("\tmovl\t$%d, %%e%cx", value, reg[0]);
  printf("\t#pushcharconst '%c'\n", value);
}

pushstrconst() {
  printf("\t\t#pushstrconst %s\n", ident);
  printf("\t.section .rodata\n");
  printf(".LC%d:\n", lco);
  printf("\t.string\t%s\n", ident);
  printf("\t.text\n");
  push();
  printf("\tmovl\t$.LC%d, %%e%cx\n", lco, reg[0]);
  ++lco;
}

pushvar() {
  // type dependent
  push();
  printf("\tmovl\t%s, %%e%cx", ident, reg[0]);
  printf("\t#pushintvar %s\n", ident);
}

pushchrvar(int ident) {
  push();
  printf("\tmovbsl\t%c, %%e%cx", ident, reg[0]);
  printf("\t#pushchrvar %c\n", ident);
}

pusharrayvar() {
  // type dependent
  printf("\tmovl\t%s(,%%e%cx,4), %%e%cx", idarray, reg[0], reg[0]);
  printf("\t#pusharrayintvar %s\n", idarray);
}

pushchrarrayvar() {
  // type dependent
  printf("\tmovl\t%s(,%%e%cx,1), %%e%cx", idarray, reg[0], reg[0]);
  printf("\t#pushchrarrayintvar %s\n", idarray);
}

add() {
  printf("\taddl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

sub() {
  printf("\tsubl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

mul() {
  printf("\timul\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}  

div() {
  int tmp;
  if ('a' == reg[0]) {
    printf("\txchgl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
    tmp = reg[0]; 
    reg[0] = reg[1];
    reg[1] = tmp;
  }
  if ('a' == reg[2]) {
    printf("\txchgl\t%%e%cx, %%e%cx\n", reg[2], reg[1]);
    tmp = reg[1]; 
    reg[1] = reg[2];
    reg[2] = tmp;
  }
  printf("\tcltd\n");
  printf("\tidivl\t%%e%cx\n", reg[0]);
  pop();
} 
 
mod() {
  div();
  printf("\tmovl\t%%edx, %%eax\n");  
}

neg() {
  printf("\tnegl\t%%e%cx\n", reg[0]);
}

and() {
  printf("\tandl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

or() {
  printf("\torl\t%%e%cx, %%e%cx\n", reg[0], reg[1]);
  pop();
}

not() {
  printf("\tnotl\t%%e%cx\n", reg[0]);
}

eq() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#equal\n");
  printf("\tsete\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
ne() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#not equal\n");
  printf("\tsetne\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
lt() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#less\n");
  printf("\tsetl\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
le() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#less or equal\n");
  printf("\tsetle\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
gt() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#greater\n");
  printf("\tsetg\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  
ge() {
  printf("\tcmpl\t%%e%cx, %%e%cx", reg[0], reg[1]);
  printf("\t#greater or equal\n");
  printf("\tsetge\t%%%cl\n", reg[1]);
  printf("\tmovzbl\t%%%cl, %%e%cx\n",reg[1], reg[1]);
  pop();
}  
  

#ifdef MODULTEST
main() {
  // c = a - b
  pushvar('a');
  pushvar('b');
  sub();
  pop('c');
  
  pushvar('a');
  pushvar('b');
  pushvar('c');
  pushvar('d');
  pushvar('e');
  pushvar('f');
  sub();
  pop('x');
  sub();
  pop('y');
  sub();
  pop('z');
  
  // a = a - 1
  pushvar('a');
  pushvar('1');
  sub();
  pop('a');
}  
#endif   
